@@ -16,7 +16,7 @@ from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
 from trytond import backend
 from trytond.pyson import If, Eval, Bool, PYSONEncoder, Id
 from trytond.transaction import Transaction
-from trytond.pool import Pool, PoolMeta
+from trytond.pool import Pool
 
 from trytond.modules.account.tax import TaxableMixin
 from trytond.modules.product import price_digits
@@ -27,7 +27,6 @@ __all__ = ['Sale', 'SaleIgnoredInvoice', 'SaleRecreatedInvoice',
     'HandleShipmentExceptionAsk', 'HandleShipmentException',
     'HandleInvoiceExceptionAsk', 'HandleInvoiceException',
     'ReturnSaleStart', 'ReturnSale']
-__metaclass__ = PoolMeta
 
 _ZERO = Decimal(0)
 
@@ -35,7 +34,7 @@ _ZERO = Decimal(0)
 class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     'Sale'
     __name__ = 'sale.sale'
-    _rec_name = 'reference'
+    _rec_name = 'number'
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
@@ -45,7 +44,8 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                 Eval('context', {}).get('company', -1)),
             ],
         depends=['state'], select=True)
-    reference = fields.Char('Reference', readonly=True, select=True)
+    number = fields.Char('Number', readonly=True, select=True)
+    reference = fields.Char('Reference', select=True)
     description = fields.Char('Description',
         states={
             'readonly': Eval('state') != 'draft',
@@ -260,7 +260,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         TableHandler = backend.get('TableHandler')
         sale_line_invoice_line_table_name = 'sale_line_invoice_lines_rel'
         Move = pool.get('stock.move')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         model_data = Table('ir_model_data')
         model_field = Table('ir_model_field')
         sql_table = cls.__table__()
@@ -287,11 +287,16 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                         len('packing'))],
                 where=model_field.name.like('%packing%')
                 & (model_field.module == module_name)))
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.column_rename('packing_state', 'shipment_state')
         table.column_rename('packing_method', 'shipment_method')
         table.column_rename('packing_address', 'shipment_address')
 
+        # Migration from 3.8: rename reference into number
+        if (table.column_exist('reference')
+                and not table.column_exist('number')):
+            table.column_rename('reference', 'number')
+
         super(Sale, cls).__register__(module_name)
 
         # Migration from 1.2
@@ -300,15 +305,15 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                 values=['shipment'],
                 where=sql_table.invoice_method == 'packing'))
 
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         # Migration from 2.2
         table.not_null_action('sale_date', 'remove')
 
         # state confirmed splitted into confirmed and processing
-        if (TableHandler.table_exist(cursor, SaleLine._table)
-                and TableHandler.table_exist(cursor,
+        if (TableHandler.table_exist(SaleLine._table)
+                and TableHandler.table_exist(
                     sale_line_invoice_line_table_name)
-                and TableHandler.table_exist(cursor, Move._table)):
+                and TableHandler.table_exist(Move._table)):
             sale_line = SaleLine.__table__()
             sale_line_invoice_line = \
                 Table(sale_line_invoice_line_table_name)
@@ -333,7 +338,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                     where=sql_table.id.in_(sub_query.select(sub_query.id))))
 
         # Add index on create_date
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @classmethod
@@ -470,10 +475,6 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                     if value is not None else default_value,)
         return taxable_lines
 
-    @property
-    def tax_type(self):
-        return 'invoice'
-
     def get_tax_amount(self):
         return sum(v['amount'] for v in self._get_taxes().itervalues())
 
@@ -577,7 +578,11 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         Search on shipments or returns
         '''
         def method(self, name, clause):
-            return [('lines.moves.shipment',) + tuple(clause[1:])
+            if isinstance(clause[2], basestring):
+                target = 'rec_name'
+            else:
+                target = 'id'
+            return [('lines.moves.shipment.' + target,) + tuple(clause[1:])
                 + (model_name,)]
         return classmethod(method)
 
@@ -640,20 +645,24 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             self.raise_user_error('invalid_method', (self.rec_name,))
 
     def get_rec_name(self, name):
-        return (self.reference or str(self.id)
+        return (self.number or str(self.id)
             + ' - ' + self.party.rec_name)
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        if clause[1].startswith('!') or clause[1].startswith('not '):
+        _, operator, value = clause
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        names = clause[2].split(' - ', 1)
-        res = [bool_op, ('reference', clause[1], names[0])]
+        names = value.split(' - ', 1)
+        domain = [bool_op,
+            ('number', operator, names[0]),
+            ('reference', operator, names[0]),
+            ]
         if len(names) != 1 and names[1]:
-            res.append(('party', clause[1], names[1]))
-        return res
+            domain = [bool_op, domain, ('party', operator, names[1])]
+        return domain
 
     @classmethod
     def view_attributes(cls):
@@ -665,7 +674,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             default = {}
         default = default.copy()
         default['state'] = 'draft'
-        default['reference'] = None
+        default['number'] = None
         default['invoice_state'] = 'none'
         default['invoices_ignored'] = None
         default['moves'] = None
@@ -677,7 +686,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         if not self.invoice_address or not self.shipment_address:
             self.raise_user_error('addresses_required', (self.rec_name,))
         for line in self.lines:
-            if line.quantity >= 0:
+            if (line.quantity or 0) >= 0:
                 location = line.from_location
             else:
                 location = line.to_location
@@ -688,9 +697,9 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                     (self.rec_name,))
 
     @classmethod
-    def set_reference(cls, sales):
+    def set_number(cls, sales):
         '''
-        Fill the reference field with the sale sequence
+        Fill the number field with the sale sequence
         '''
         pool = Pool()
         Sequence = pool.get('ir.sequence')
@@ -698,12 +707,10 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
 
         config = Config(1)
         for sale in sales:
-            if sale.reference:
+            if sale.number:
                 continue
-            reference = Sequence.get_id(config.sale_sequence.id)
-            cls.write([sale], {
-                    'reference': reference,
-                    })
+            sale.number = Sequence.get_id(config.sale_sequence.id)
+        cls.save(sales)
 
     @classmethod
     def set_sale_date(cls, sales):
@@ -723,48 +730,24 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
                     'total_amount_cache': sale.total_amount,
                     })
 
-    def _get_invoice_line_sale_line(self, invoice_type):
-        '''
-        Return invoice line for each sale lines according to invoice_type
-        '''
-        res = {}
-        for line in self.lines:
-            val = line.get_invoice_line(invoice_type)
-            if val:
-                res[line.id] = val
-        return res
-
-    def _get_invoice_sale(self, invoice_type):
-        '''
-        Return invoice of type invoice_type
-        '''
+    def _get_invoice_sale(self):
+        'Return invoice'
         pool = Pool()
         Invoice = pool.get('account.invoice')
-        Journal = pool.get('account.journal')
-
-        journals = Journal.search([
-                ('type', '=', 'revenue'),
-                ], limit=1)
-        if journals:
-            journal, = journals
-        else:
-            journal = None
-
-        return Invoice(
+        invoice = Invoice(
             company=self.company,
-            type=invoice_type,
-            journal=journal,
+            type='out',
             party=self.party,
             invoice_address=self.invoice_address,
             currency=self.currency,
             account=self.party.account_receivable,
-            payment_term=self.payment_term,
             )
+        invoice.on_change_type()
+        invoice.payment_term = self.payment_term
+        return invoice
 
-    def create_invoice(self, invoice_type):
-        '''
-        Create and return an invoice of type invoice_type
-        '''
+    def create_invoice(self):
+        'Create and return an invoice'
         pool = Pool()
         Invoice = pool.get('account.invoice')
         if self.invoice_method == 'manual':
@@ -774,11 +757,15 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
             self.raise_user_error('missing_account_receivable',
                 (self.party.rec_name,))
 
-        invoice_lines = self._get_invoice_line_sale_line(invoice_type)
+        invoice_lines = {}
+        for line in self.lines:
+            ilines = line.get_invoice_line()
+            if ilines:
+                invoice_lines[line.id] = ilines
         if not invoice_lines:
             return
 
-        invoice = self._get_invoice_sale(invoice_type)
+        invoice = self._get_invoice_sale()
         invoice.lines = ((list(invoice.lines)
                 if hasattr(invoice, 'lines') else [])
             + list(chain.from_iterable(invoice_lines[l.id] for l in self.lines
@@ -788,17 +775,6 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         Invoice.update_taxes([invoice])
         return invoice
 
-    def _get_move_sale_line(self, shipment_type):
-        '''
-        Return move for each sale lines of the right shipment_type
-        '''
-        res = {}
-        for line in self.lines:
-            val = line.get_move(shipment_type)
-            if val:
-                res[line.id] = val
-        return res
-
     def _group_shipment_key(self, moves, move):
         '''
         The key to group moves by shipments
@@ -836,7 +812,11 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         if self.shipment_method == 'manual':
             return
 
-        moves = self._get_move_sale_line(shipment_type)
+        moves = {}
+        for line in self.lines:
+            move = line.get_move(shipment_type)
+            if move:
+                moves[line.id] = move
         if not moves:
             return
         if shipment_type == 'out':
@@ -894,7 +874,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
     def quote(cls, sales):
         for sale in sales:
             sale.check_for_quotation()
-        cls.set_reference(sales)
+        cls.set_number(sales)
 
     @classmethod
     @ModelView.button
@@ -931,8 +911,7 @@ class Sale(Workflow, ModelSQL, ModelView, TaxableMixin):
         for sale in sales:
             if sale.state not in ('confirmed', 'processing', 'done'):
                 continue
-            sale.create_invoice('out_invoice')
-            sale.create_invoice('out_credit_note')
+            sale.create_invoice()
             sale.set_invoice_state()
             sale.create_shipment('out')
             sale.create_shipment('return')
@@ -989,7 +968,7 @@ class SaleLine(ModelSQL, ModelView):
             'readonly': ~Eval('_parent_sale', {}),
             },
         depends=['type', 'unit_digits'])
-    unit = fields.Many2One('product.uom', 'Unit',
+    unit = fields.Many2One('product.uom', 'Unit', ondelete='RESTRICT',
             states={
                 'required': Bool(Eval('product')),
                 'invisible': Eval('type') != 'line',
@@ -1004,7 +983,7 @@ class SaleLine(ModelSQL, ModelView):
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
         'on_change_with_unit_digits')
     product = fields.Many2One('product.product', 'Product',
-        domain=[('salable', '=', True)],
+        ondelete='RESTRICT', domain=[('salable', '=', True)],
         states={
             'invisible': Eval('type') != 'line',
             'readonly': ~Eval('_parent_sale', {}),
@@ -1059,12 +1038,12 @@ class SaleLine(ModelSQL, ModelView):
             'From Location'), 'get_from_location')
     to_location = fields.Function(fields.Many2One('stock.location',
             'To Location'), 'get_to_location')
-    delivery_date = fields.Function(fields.Date('Delivery Date',
+    shipping_date = fields.Function(fields.Date('Shipping Date',
             states={
                 'invisible': Eval('type') != 'line',
                 },
             depends=['type']),
-        'on_change_with_delivery_date')
+        'on_change_with_shipping_date')
 
     @classmethod
     def __setup__(cls):
@@ -1083,10 +1062,10 @@ class SaleLine(ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         TableHandler = backend.get('TableHandler')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         sql_table = cls.__table__()
         super(SaleLine, cls).__register__(module_name)
-        table = TableHandler(cursor, cls, module_name)
+        table = TableHandler(cls, module_name)
 
         # Migration from 1.0 comment change into note
         if table.column_exist('comment'):
@@ -1282,34 +1261,35 @@ class SaleLine(ModelSQL, ModelView):
         return self.sale.warehouse.id if self.sale.warehouse else None
 
     def get_from_location(self, name):
-        if self.quantity >= 0:
+        if (self.quantity or 0) >= 0:
             if self.warehouse:
                 return self.warehouse.output_location.id
         else:
             return self.sale.party.customer_location.id
 
     def get_to_location(self, name):
-        if self.quantity >= 0:
+        if (self.quantity or 0) >= 0:
             return self.sale.party.customer_location.id
         else:
             if self.warehouse:
                 return self.warehouse.input_location.id
 
     @fields.depends('product', 'quantity', 'moves', '_parent_sale.sale_date')
-    def on_change_with_delivery_date(self, name=None):
+    def on_change_with_shipping_date(self, name=None):
         if self.moves:
-            return min(m.effective_date if m.effective_date else m.planned_date
-                for m in self.moves)
-        if self.product and self.quantity > 0:
+            dates = filter(
+                None, (m.effective_date or m.planned_date for m in self.moves))
+            if dates:
+                return min(dates)
+            else:
+                return
+        if self.product and self.quantity is not None and self.quantity > 0:
             date = self.sale.sale_date if self.sale else None
-            return self.product.compute_delivery_date(date=date)
+            return self.product.compute_shipping_date(date=date)
 
-    def get_invoice_line(self, invoice_type):
-        '''
-        Return a list of invoice lines for sale line according to invoice_type
-        '''
+    def get_invoice_line(self):
+        'Return a list of invoice lines for sale line'
         pool = Pool()
-        Uom = pool.get('product.uom')
         Property = pool.get('ir.property')
         InvoiceLine = pool.get('account.invoice.line')
 
@@ -1319,64 +1299,26 @@ class SaleLine(ModelSQL, ModelView):
         invoice_line.note = self.note
         invoice_line.origin = self
         if self.type != 'line':
-            if (self.sale.invoice_method == 'order'
-                    and not self.invoice_lines
-                    and ((all(l.quantity >= 0 for l in self.sale.lines
-                                if l.type == 'line')
-                            and invoice_type == 'out_invoice')
-                        or (all(l.quantity <= 0 for l in self.sale.lines
-                                if l.type == 'line')
-                            and invoice_type == 'out_credit_note'))):
+            if self._get_invoice_not_line():
                 return [invoice_line]
             else:
                 return []
 
-        if (invoice_type == 'out_invoice') != (self.quantity >= 0):
-            return []
+        quantity = (self._get_invoice_line_quantity()
+            - self._get_invoiced_quantity())
 
-        stock_moves = []
-        if (self.sale.invoice_method == 'order'
-                or not self.product
-                or self.product.type == 'service'):
-            quantity = abs(self.quantity)
-            stock_moves = self.moves
-        else:
-            quantity = 0.0
-            for move in self.moves:
-                if move.state == 'done':
-                    quantity += Uom.compute_qty(move.uom, move.quantity,
-                        self.unit)
-                    if move.invoiced_quantity < move.quantity:
-                        stock_moves.append(move)
-        if self.sale.shipment_method == 'order':
-            invoice_line.stock_moves = stock_moves
+        if self.unit:
+            quantity = self.unit.round(quantity)
+        invoice_line.quantity = quantity
 
-        skip_ids = set(l.id for i in self.sale.invoices_recreated
-            for l in i.lines)
-        for old_invoice_line in self.invoice_lines:
-            if old_invoice_line.type != 'line':
-                continue
-            if old_invoice_line.id not in skip_ids:
-                if old_invoice_line.invoice:
-                    old_invoice_type = old_invoice_line.invoice.type
-                else:
-                    old_invoice_type = old_invoice_line.invoice_type
-
-                sign = 1 if old_invoice_type == invoice_type else -1
-
-                quantity -= Uom.compute_qty(old_invoice_line.unit,
-                    sign * old_invoice_line.quantity, self.unit)
-
-        rounding = self.unit.rounding if self.unit else 0.01
-        invoice_line.quantity = Uom.round(quantity, rounding)
-        if invoice_line.quantity <= 0:
+        if not invoice_line.quantity:
             return []
 
         invoice_line.unit = self.unit
         invoice_line.product = self.product
         invoice_line.unit_price = self.unit_price
         invoice_line.taxes = self.taxes
-        invoice_line.invoice_type = invoice_type
+        invoice_line.invoice_type = 'out'
         if self.product:
             invoice_line.account = self.product.account_revenue_used
             if not invoice_line.account:
@@ -1393,31 +1335,65 @@ class SaleLine(ModelSQL, ModelView):
                 self.raise_user_error('missing_account_revenue_property', {
                         'sale': self.sale.rec_name,
                         })
+        invoice_line.stock_moves = self._get_invoice_line_moves()
         return [invoice_line]
 
-    @classmethod
-    def view_attributes(cls):
-        return [
-            ('/form//field[@name="note"]|/form//field[@name="description"]',
-                'spell', Eval('_parent_sale', {}).get('party_lang'))]
+    def _get_invoice_not_line(self):
+        'Return if the not line should be invoiced'
+        return self.sale.invoice_method == 'order' and not self.invoice_lines
 
-    @classmethod
-    def copy(cls, lines, default=None):
-        if default is None:
-            default = {}
-        default = default.copy()
-        default['moves'] = None
-        default['moves_ignored'] = None
-        default['moves_recreated'] = None
-        default['invoice_lines'] = None
-        return super(SaleLine, cls).copy(lines, default=default)
+    def _get_invoice_line_quantity(self):
+        'Return the quantity that should be invoiced'
+        pool = Pool()
+        Uom = pool.get('product.uom')
+
+        if (self.sale.invoice_method == 'order'
+                or not self.product
+                or self.product.type == 'service'):
+            return self.quantity
+        elif self.sale.invoice_method == 'shipment':
+            quantity = 0.0
+            for move in self.moves:
+                if move.state == 'done':
+                    quantity += Uom.compute_qty(move.uom, move.quantity,
+                        self.unit)
+            if self.quantity < 0:
+                quantity *= -1
+            return quantity
+
+    def _get_invoiced_quantity(self):
+        'Return the quantity already invoiced'
+        pool = Pool()
+        Uom = pool.get('product.uom')
+
+        quantity = 0
+        skips = {l for i in self.sale.invoices_recreated for l in i.lines}
+        for invoice_line in self.invoice_lines:
+            if invoice_line.type != 'line':
+                continue
+            if invoice_line not in skips:
+                quantity += Uom.compute_qty(invoice_line.unit,
+                    invoice_line.quantity, self.unit)
+        return quantity
+
+    def _get_invoice_line_moves(self):
+        'Return the stock moves that should be invoiced'
+        moves = []
+        if self.sale.invoice_method == 'order':
+            if self.sale.shipment_method != 'order':
+                moves.extend(self.moves)
+        elif self.sale.invoice_method == 'shipment':
+            for move in self.moves:
+                if move.state == 'done':
+                    if move.invoiced_quantity < move.quantity:
+                        moves.append(move)
+        return moves
 
     def get_move(self, shipment_type):
         '''
         Return moves for the sale line according to shipment_type
         '''
         pool = Pool()
-        Uom = pool.get('product.uom')
         Move = pool.get('stock.move')
 
         if self.type != 'line':
@@ -1426,30 +1402,14 @@ class SaleLine(ModelSQL, ModelView):
             return
         if self.product.type == 'service':
             return
+
         if (shipment_type == 'out') != (self.quantity >= 0):
             return
 
-        invoice_lines = []
-        if self.sale.shipment_method == 'order':
-            quantity = abs(self.quantity)
-            if self.sale.invoice_method == 'order':
-                invoice_lines = self.invoice_lines
-        else:
-            quantity = 0.0
-            for invoice_line in self.invoice_lines:
-                if invoice_line.invoice.state == 'paid':
-                    quantity += Uom.compute_qty(invoice_line.unit,
-                        invoice_line.quantity, self.unit)
-                    if invoice_line.moved_quantity < invoice_line.quantity:
-                        invoice_lines.append(invoice_line)
+        quantity = (self._get_move_quantity(shipment_type)
+            - self._get_shipped_quantity(shipment_type))
 
-        skip_ids = set(x.id for x in self.moves_recreated)
-        for move in self.moves:
-            if move.id not in skip_ids:
-                quantity -= Uom.compute_qty(move.uom, move.quantity,
-                    self.unit)
-
-        quantity = Uom.round(quantity, self.unit.rounding)
+        quantity = self.unit.round(quantity)
         if quantity <= 0:
             return
 
@@ -1468,11 +1428,71 @@ class SaleLine(ModelSQL, ModelView):
         move.company = self.sale.company
         move.unit_price = self.unit_price
         move.currency = self.sale.currency
-        move.planned_date = self.delivery_date
-        move.invoice_lines = invoice_lines
+        move.planned_date = self.shipping_date
+        move.invoice_lines = self._get_move_invoice_lines(shipment_type)
         move.origin = self
         return move
 
+    def _get_move_quantity(self, shipment_type):
+        'Return the quantity that should be shipped'
+        pool = Pool()
+        Uom = pool.get('product.uom')
+
+        if self.sale.shipment_method == 'order':
+            return abs(self.quantity)
+        elif self.sale.shipment_method == 'invoice':
+            quantity = 0.0
+            for invoice_line in self.invoice_lines:
+                if (invoice_line.invoice
+                        and invoice_line.invoice.state == 'paid'):
+                    quantity += Uom.compute_qty(invoice_line.unit,
+                        invoice_line.quantity, self.unit)
+            return quantity
+
+    def _get_shipped_quantity(self, shipment_type):
+        'Return the quantity already shipped'
+        pool = Pool()
+        Uom = pool.get('product.uom')
+
+        quantity = 0
+        skips = set(m for m in self.moves_recreated)
+        for move in self.moves:
+            if move not in skips:
+                quantity += Uom.compute_qty(move.uom, move.quantity,
+                    self.unit)
+        return quantity
+
+    def _get_move_invoice_lines(self, shipment_type):
+        'Return the invoice lines that should be shipped'
+        invoice_lines = []
+        if self.sale.shipment_method == 'order':
+            if self.sale.invoice_method == 'order':
+                invoice_lines.extend(self.invoice_lines)
+        elif self.sale.shipment_method == 'invoice':
+            for invoice_line in self.invoice_lines:
+                if (invoice_line.invoice
+                        and invoice_line.invoice.state == 'paid'):
+                    if invoice_line.moved_quantity < invoice_line.quantity:
+                        invoice_lines.append(invoice_line)
+        return invoice_lines
+
+    @classmethod
+    def view_attributes(cls):
+        return [
+            ('/form//field[@name="note"]|/form//field[@name="description"]',
+                'spell', Eval('_parent_sale', {}).get('party_lang'))]
+
+    @classmethod
+    def copy(cls, lines, default=None):
+        if default is None:
+            default = {}
+        default = default.copy()
+        default['moves'] = None
+        default['moves_ignored'] = None
+        default['moves_recreated'] = None
+        default['invoice_lines'] = None
+        return super(SaleLine, cls).copy(lines, default=default)
+
 
 class SaleLineTax(ModelSQL):
     'Sale Line - Tax'
@@ -1519,11 +1539,11 @@ class OpenCustomer(Wizard):
         ModelData = pool.get('ir.model.data')
         Wizard = pool.get('ir.action.wizard')
         Sale = pool.get('sale.sale')
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         sale = Sale.__table__()
 
         cursor.execute(*sale.select(sale.party, group_by=sale.party))
-        customer_ids = [line[0] for line in Transaction().cursor.fetchall()]
+        customer_ids = [line[0] for line in cursor.fetchall()]
         action['pyson_domain'] = PYSONEncoder().encode(
             [('id', 'in', customer_ids)])
         wizard = Wizard(ModelData.get_id('sale', 'act_open_customer'))
@@ -1545,7 +1565,7 @@ class HandleShipmentExceptionAsk(ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         model = Table('ir_model')
         # Migration from 1.2: packing renamed into shipment
         cursor.execute(*model.update(
